/* kprobe_post_handler_work.c
*/

#include <asm/timex.h>
#include <linux/module.h>
#include <linux/kprobes.h>
#include <linux/fs.h>

#define RUNS 3
#define LOOPS 100000

/*For each probe you need to allocate a kprobe structure*/
static struct kprobe kp;

static int kprobe_triggers=0;
static int kprobe_pre_triggers=0;
static int kprobe_post_triggers=0;

/*kprobe pre_handler: called just before the probed instruction is executed*/

int handler_pre(struct kprobe *p, struct pt_regs *regs)
{
	++kprobe_pre_triggers;
	return 0;
}

/*kprobe post_handler: called after the probed instruction is executed*/

void handler_post(struct kprobe *p, struct pt_regs *regs, unsigned long flags)
{
	++kprobe_post_triggers;
}

/* fault_handler: this is called if an exception is generated for any
 * instruction within the fault-handler, or when Kprobes
 * single-steps the probed instruction.
 */
int handler_fault(struct kprobe *p, struct pt_regs *regs, int trapnr)
{
	/* do nothing */
	return 0;
}

extern void do_nothing(void);

static int run_test(void)
{
  int i, j;
  cycles_t start, stop;
  unsigned int result, clocks_per_iter;
  unsigned int average_clocks = 0;

  for (j=0; j<RUNS; ++j){
	  kprobe_triggers=0;
	  kprobe_pre_triggers=0;
	  kprobe_post_triggers=0;
	  /* estimate kprobe */
	  start = get_cycles();
	  for (i=0; i<LOOPS; ++i){
		  do_nothing();
	  }
	  stop=get_cycles();
	  result = (stop-start);
	  clocks_per_iter = result/kprobe_triggers;
	  printk("kprobe start %lld, stop %lld\n", start, stop);
	  printk("kprobe overhead count %d, number of cycles %d\n",
		 kprobe_triggers, result);
	  printk("kprobe_pre_trigger %d\n", kprobe_pre_triggers);
	  printk("kprobe_post_trigger %d\n", kprobe_post_triggers);
	  printk("kprobe overhead of cycles %d per iteration\n",
		 clocks_per_iter);
	  average_clocks += clocks_per_iter;
  }
  average_clocks = average_clocks / RUNS;
  return average_clocks;
}

int init_module(void)
{
	int result;

	kp.pre_handler = handler_pre;
	kp.post_handler = handler_post;
	kp.fault_handler = handler_fault;
	kp.addr = (kprobe_opcode_t *) do_nothing;
	/* register the kprobe now */
	result = register_kprobe(&kp);
	if (result) {
		printk("Unable to register probe.\n");
		return result;
	}
	printk("kprobe registered\n");

	return 0;
}

void cleanup_module(void)
{
	int result;

	result = run_test();

	unregister_kprobe(&kp);
	printk("kprobe unregistered\n");
}

void do_nothing()
{
	++kprobe_triggers;
}

MODULE_LICENSE("GPL");
